"""
Test the importing module.
This module relies heavily on an external service and requires
quite a bit of mocking.
"""
import os
from django.test import TestCase
from mock import patch

from molo.core.api import importers
from molo.core.api.tests import constants, utils
from molo.core.models import (
    ArticlePage,
    SectionPage,
    FooterPage,
    BannerPage,
    Tag,
    ArticlePageRecommendedSections,
    ArticlePageRelatedSections,
    SiteLanguageRelation,
    ArticlePageTags,
    SectionPageTags,
)
from molo.core.tests.base import MoloTestCaseMixin

import responses

from wagtail.wagtailcore.models import Site
from wagtail.wagtailimages.tests.utils import Image, get_test_image_file

from dateutil import parser


class ArticleImportTestCase(MoloTestCaseMixin, TestCase):

    def setUp(self):
        self.mk_main()
        self.importer = importers.ArticlePageImporter(
            base_url="http://localhost:8000",
            content=constants.AVAILABLE_ARTICLES["items"]
        )

    def test_article_importer_initializtion(self):
        self.assertEqual(
            self.importer.content(),
            constants.AVAILABLE_ARTICLES["items"]
        )

    @patch("molo.core.api.importers.get_image")
    def test_articles_can_be_saved(self, mock_image):
        image = Image.objects.create(
            title="Test image",
            file=get_test_image_file(),
        )
        mock_image.return_value = image

        # Create parent page to which articles will be saved
        section = self.mk_section(
            self.section_index, title="Parent Test Section 2",
        )
        self.assertEqual(ArticlePage.objects.all().count(), 0)

        # Save the articles
        # Save the first available article
        # import pdb;pdb.set_trace()
        self.importer.save([0, ], section.id)
        self.assertEqual(ArticlePage.objects.all().count(), 1)

    def test_nested_fields_can_be_extracted(self):
        # It is necessary to separate nested fields in each article
        # dictionary from those that are not nested. Reason being
        # some of the nested fields have to be treated differently.
        pass

    @patch("molo.core.api.importers.get_image")
    def test_related_image_can_be_retrieved(self, mock_image):
        mock_image.return_value = constants.RELATED_IMAGE
        # assert 0
        # self.assertIsInstance(
        #     importers.get_image(base_url=self.importer.base_url, image_id=1),
        #     Image
        # )

    def tearDown(self):
        del self.importer


class SectionImportTestCase(MoloTestCaseMixin, TestCase):

    def setUp(self):
        self.mk_main()
        self.importer = importers.SectionPageImporter(
            base_url="http://localhost:8000",
            content=constants.AVAILABLE_SECTIONS["items"]
        )

    def test_section_importer_initializtion(self):
        self.assertEqual(
            self.importer.content(),
            constants.AVAILABLE_SECTIONS["items"]
        )

    def tearDown(self):
        del self.importer

    # @patch("molo.core.api.importers.get_image")
    # def test_section_can_be_saved(self, mock_image):
    #     image = Image.objects.create(
    #         title="Test image",
    #         file=get_test_image_file(),
    #     )
    #     mock_image.return_value = image
    #
    #     # Add new sections as children pf the SectionIndexPage
    #     self.assertEqual(SectionPage.objects.all().count(), 0)
    #
    #     # Save the articles
    #     # Save the first available article
    #     self.importer.save([0, ], self.section_index.id)
    #     self.assertEqual(SectionPage.objects.all().count(), 1)


class TestImporterUtilFunctions(TestCase):
    def setUp(self):
        self.test_url = "http://localhost:8000/api/v2/images/"

    @responses.activate
    def test_list_of_objects_from_api(self):
        responses.add(responses.GET,
                      self.test_url,
                      json=constants.WAGTAIL_API_LIST_VIEW, status=200)
        returned_list = importers.list_of_objects_from_api(self.test_url)
        self.assertEqual(
            returned_list,
            constants.WAGTAIL_API_LIST_VIEW["items"])

    @patch("molo.core.api.importers.requests.get",
           side_effect=utils.mocked_requests_get)
    def test_list_of_objects_from_api_paginated(self, mock_get):
        responses.add(responses.GET,
                      self.test_url,
                      json=constants.WAGTAIL_API_LIST_VIEW_PAGE_1, status=200)
        responses.add(responses.GET,
                      "{}?limit=20&offset=20".format(self.test_url),
                      json=constants.WAGTAIL_API_LIST_VIEW_PAGE_2, status=200)
        returned_list = importers.list_of_objects_from_api(self.test_url)
        expected_response = (
            constants.WAGTAIL_API_LIST_VIEW_PAGE_1["items"] +
            constants.WAGTAIL_API_LIST_VIEW_PAGE_2["items"]
        )
        self.assertEqual(
            returned_list,
            expected_response)


class TestSiteSectionImporter(MoloTestCaseMixin, TestCase):
    def setUp(self):
        self.fake_base_url = "http://localhost:8000"
        self.mk_main()
        self.importer = importers.SiteImporter(self.site.pk,
                                               self.fake_base_url)

    @responses.activate
    def test_get_language_ids(self):
        responses.add(responses.GET,
                      "{}/api/v2/languages/".format(self.fake_base_url),
                      json=constants.LANGUAGE_LIST_RESPONSE, status=200)
        self.assertEqual(
            self.importer.get_language_ids(),
            [constants.LANGUAGE_LIST_RESPONSE["items"][0]["id"],
             constants.LANGUAGE_LIST_RESPONSE["items"][1]["id"]])

    @responses.activate
    def test_copy_site_languages(self):
        responses.add(responses.GET,
                      "{}/api/v2/languages/".format(self.fake_base_url),
                      json=constants.LANGUAGE_LIST_RESPONSE, status=200)
        responses.add(responses.GET,
                      "{}/api/v2/languages/1/".format(self.fake_base_url),
                      json=constants.LANGUAGE_RESPONSE_1, status=200)
        responses.add(responses.GET,
                      "{}/api/v2/languages/2/".format(self.fake_base_url),
                      json=constants.LANGUAGE_RESPONSE_2, status=200)

        self.importer.copy_site_languages()

        eng_lang = SiteLanguageRelation.objects.get(locale="en")
        fr_lang = SiteLanguageRelation.objects.get(locale="fr")

        self.assertTrue(eng_lang)
        self.assertTrue(eng_lang.is_active)
        self.assertTrue(eng_lang.is_main_language)

        self.assertTrue(fr_lang)
        self.assertTrue(fr_lang.is_active)
        self.assertFalse(fr_lang.is_main_language)

    @responses.activate
    def test_fetch_and_create_image_new_image(self):
        image_title = "test_title.png"
        relative_url = "/media/images/SIbomiWV1AQ.original.jpg"
        test_file_path = os.getcwd() + '/molo/core/api/tests/test_image.png'

        with open(test_file_path, 'rb') as img1:
            responses.add(
                responses.GET, '{}{}'.format(self.fake_base_url, relative_url),
                body=img1.read(), status=200,
                content_type='image/jpeg',
                stream=True
            )
        result = self.importer.fetch_and_create_image(
            relative_url,
            image_title)

        self.assertEqual(type(result), Image)
        self.assertEqual(result.title, image_title)
        self.assertEqual(Image.objects.count(), 1)

    @responses.activate
    @patch("molo.core.api.importers.SiteImporter.fetch_and_create_image",
           side_effect=utils.mocked_fetch_and_create_image)
    def test_import_images(self, mock_fetch_and_create_image):
        image_list_url = '{}/api/v2/images/'.format(self.fake_base_url)
        image_detail_url_1 = "{}{}/".format(image_list_url,
                                            constants.IMAGE_DETAIL_1["id"])
        image_detail_url_2 = "{}{}/".format(image_list_url,
                                            constants.IMAGE_DETAIL_2["id"])
        responses.add(
            responses.GET, image_list_url,
            json=constants.IMAGE_LIST_RESPONSE, status=200)
        responses.add(
            responses.GET, image_detail_url_1,
            json=constants.IMAGE_DETAIL_1, status=200)
        responses.add(
            responses.GET, image_detail_url_2,
            json=constants.IMAGE_DETAIL_2, status=200)

        self.assertEqual(Image.objects.count(), 0)

        self.importer.import_images()

        self.assertEqual(Image.objects.count(), 2)
        self.assertEqual(
            Image.objects.first().title,
            constants.IMAGE_DETAIL_1["title"])
        self.assertEqual(
            Image.objects.last().title,
            constants.IMAGE_DETAIL_2["title"])

        # check that image mapping from foreign to local exists
        self.assertEqual(
            self.importer.image_map[constants.IMAGE_DETAIL_1["id"]],
            Image.objects.first().id)
        self.assertEqual(
            self.importer.image_map[constants.IMAGE_DETAIL_2["id"]],
            Image.objects.last().id)

    @responses.activate
    @patch("molo.core.api.importers.SiteImporter.fetch_and_create_image",
           side_effect=utils.mocked_fetch_and_create_image)
    def test_import_images_avoid_duplicates(self, mock_fetch_and_create_image):
        image_list_url = '{}/api/v2/images/'.format(self.fake_base_url)
        image_detail_url_1 = "{}{}/".format(image_list_url,
                                            constants.IMAGE_DETAIL_1["id"])
        image_detail_url_2 = "{}{}/".format(image_list_url,
                                            constants.IMAGE_DETAIL_2["id"])
        responses.add(
            responses.GET, image_list_url,
            json=constants.IMAGE_LIST_RESPONSE, status=200)
        responses.add(
            responses.GET, image_detail_url_1,
            json=constants.IMAGE_DETAIL_1, status=200)
        responses.add(
            responses.GET, image_detail_url_2,
            json=constants.IMAGE_DETAIL_2, status=200)

        # create 'duplicate' image with same name
        Image.objects.create(
            title='local image',
            file=get_test_image_file(),
        )
        self.assertEqual(Image.objects.count(), 1)

        self.importer.import_images()

        self.assertEqual(Image.objects.count(), 2)
        # Note that local title is used over foreign title
        self.assertEqual(
            Image.objects.first().title,
            'local image')
        self.assertEqual(
            Image.objects.last().title,
            constants.IMAGE_DETAIL_2["title"])

        # check that image mapping from foreign to local exists
        self.assertEqual(
            self.importer.image_map[constants.IMAGE_DETAIL_1["id"]],
            Image.objects.first().id)
        self.assertEqual(
            self.importer.image_map[constants.IMAGE_DETAIL_2["id"]],
            Image.objects.last().id)

    def check_article_and_footer_fields(self, page, content):
        self.assertEqual(page.title, content["title"])
        self.assertEqual(page.subtitle, content["subtitle"])
        self.assertEqual(page.commenting_state, content["commenting_state"])
        self.assertEqual(page.social_media_title,
                         content["social_media_title"])
        self.assertEqual(page.social_media_description,
                         content["social_media_description"])
        self.assertEqual(page.featured_in_latest,
                         content["featured_in_latest"])
        self.assertEqual(page.featured_in_section,
                         content["featured_in_section"])
        self.assertEqual(page.featured_in_homepage,
                         content["featured_in_homepage"])
        self.assertEqual(page.feature_as_topic_of_the_day,
                         content["feature_as_topic_of_the_day"])

        self.assertEqual(page.commenting_open_time,
                         parser.parse(content["commenting_open_time"]))
        self.assertEqual(page.commenting_close_time,
                         parser.parse(content["commenting_close_time"]))
        self.assertEqual(page.featured_in_latest_start_date,
                         parser.parse(
                             content["featured_in_latest_start_date"]))
        self.assertEqual(page.featured_in_latest_end_date,
                         parser.parse(content["featured_in_latest_end_date"]))
        self.assertEqual(page.featured_in_section_start_date,
                         parser.parse(
                             content["featured_in_section_start_date"]))
        self.assertEqual(page.featured_in_section_end_date,
                         parser.parse(content["featured_in_section_end_date"]))
        self.assertEqual(page.featured_in_homepage_start_date,
                         parser.parse(
                             content["featured_in_homepage_start_date"]))
        self.assertEqual(page.featured_in_homepage_end_date,
                         parser.parse(
                             content["featured_in_homepage_end_date"]))
        self.assertEqual(page.promote_date,
                         parser.parse(content["promote_date"]))
        self.assertEqual(page.demote_date,
                         parser.parse(content["demote_date"]))

        # NESTED FIELDS
        self.assertTrue(hasattr(page.body, "stream_data"))
        self.assertEqual(page.body.stream_data, content['body'])

        self.assertEqual(page.tags.count(), len(content['tags']))
        for tag in page.tags.all():
            self.assertTrue(tag.name in content["tags"])

        self.assertEqual(page.metadata_tags.count(),
                         len(content['metadata_tags']))
        for metadata_tag in page.metadata_tags.all():
            self.assertTrue(metadata_tag.name in content["metadata_tags"])

        # Check that all reference fields have been properly stored
        # for further processing later

        # nav_tags
        self.assertEqual(self.importer.id_map[content["id"]], page.id)
        self.assertTrue(page.id in self.importer.nav_tags)
        self.assertEqual(self.importer.nav_tags[page.id],
                         [content["nav_tags"][0]["tag"]["id"], ])
        # TODO
        # self.assertEqual(self.importer.image_map[page.id], [])

        self.assertTrue(page.id in self.importer.related_sections)
        self.assertEqual(
            self.importer.related_sections[page.id],
            [content["related_sections"][0]["section"]["id"],
             content["related_sections"][1]["section"]["id"]])

        self.assertTrue(page.id in self.importer.recommended_articles)
        self.assertEqual(
            self.importer.recommended_articles[page.id],
            [content["recommended_articles"][0]["recommended_article"]["id"],
             content["recommended_articles"][1]["recommended_article"]["id"]])

        self.assertTrue(page.id in self.importer.reaction_questions)
        self.assertEqual(
            self.importer.reaction_questions[page.id],
            [content["reaction_questions"][0]["reaction_question"]["id"],
             content["reaction_questions"][1]["reaction_question"]["id"]])

        # Check that image file has been added
        self.assertTrue(page.image)
        self.assertEqual(page.image.title, content["image"]["title"])
        # Check that social media file has been added
        self.assertTrue(page.social_media_image)
        self.assertEqual(page.social_media_image.title,
                         content["social_media_image"]["title"])

    def test_create_article_page(self):
        # fake the content passed to the importer
        content = utils.fake_article_page_response()
        # avoid any side effects by creating a copy of content
        content_copy = dict(content)

        # create local versions of images, mapped to foreign ID
        foreign_image_id = content["image"]["id"]
        image = Image.objects.create(
            title=content["image"]["title"],
            file=get_test_image_file(),
        )
        self.importer.image_map[foreign_image_id] = image.id

        foreign_social_media_image_id = content["social_media_image"]["id"]
        social_media_image = Image.objects.create(
            title=content["social_media_image"]["title"],
            file=get_test_image_file(),
        )

        (self.importer
             .image_map[foreign_social_media_image_id]) = social_media_image.id

        parent = self.mk_section(self.section_index)

        self.assertEqual(ArticlePage.objects.count(), 0)

        article = self.importer.create_page(parent, content_copy)

        self.assertEqual(ArticlePage.objects.count(), 1)
        self.assertEqual(article.get_parent(), parent)

        self.check_article_and_footer_fields(article, content)

    def test_create_section_page(self):
        # fake the content passed to the importer
        content = utils.fake_section_page_response()

        # create local versions of images, mapped to foreign ID
        foreign_image_id = content["image"]["id"]
        image = Image.objects.create(
            title=content["image"]["title"],
            file=get_test_image_file(),
        )
        self.importer.image_map[foreign_image_id] = image.id

        # avoid any side effects by creating a copy of content
        content_copy = dict(content)

        parent = self.section_index

        self.assertEqual(SectionPage.objects.count(), 0)

        section = self.importer.create_page(parent, content_copy)

        self.assertEqual(section.get_parent(), parent)
        self.assertNotEqual(section.id, content["id"])

        # TODO test live attribute
        self.assertEqual(section.title, content["title"])
        self.assertEqual(section.description, content["description"])
        self.assertEqual(section.extra_style_hints,
                         content["extra_style_hints"])
        self.assertEqual(section.commenting_state, content["commenting_state"])
        self.assertEqual(section.monday_rotation, content["monday_rotation"])
        self.assertEqual(section.tuesday_rotation, content["tuesday_rotation"])
        self.assertEqual(section.wednesday_rotation,
                         content["wednesday_rotation"])
        self.assertEqual(section.thursday_rotation,
                         content["thursday_rotation"])
        self.assertEqual(section.friday_rotation, content["friday_rotation"])
        self.assertEqual(section.saturday_rotation,
                         content["saturday_rotation"])
        self.assertEqual(section.sunday_rotation, content["sunday_rotation"])

        self.assertEqual(section.commenting_open_time,
                         parser.parse(content["commenting_open_time"]))
        self.assertEqual(section.commenting_close_time,
                         parser.parse(content["commenting_close_time"]))
        self.assertEqual(section.content_rotation_start_date,
                         parser.parse(content["content_rotation_start_date"]))
        self.assertEqual(section.content_rotation_end_date,
                         parser.parse(content["content_rotation_end_date"]))

        # NESTED FIELDS
        # time
        self.assertTrue(hasattr(section.time, "stream_data"))
        self.assertEqual(section.time.stream_data, content["time"])

        # section_tags/nav_tags
        self.assertEqual(self.importer.id_map[content["id"]], section.id)
        self.assertTrue(section.id in self.importer.section_tags)
        self.assertEqual(self.importer.section_tags[section.id],
                         [content["section_tags"][0]["tag"]["id"],
                          content["section_tags"][1]["tag"]["id"]])
        # Check that image file has been added
        self.assertTrue(section.image)
        self.assertEqual(section.image.title, content["image"]["title"])

    def test_create_article_page_translated(self):
        # create 2 languages
        self.english = SiteLanguageRelation.objects.create(
            language_setting=self.importer.language_setting,
            locale='en',
            is_active=True)
        self.french = SiteLanguageRelation.objects.create(
            language_setting=self.importer.language_setting,
            locale='fr',
            is_active=True)

        content = constants.ARTICLE_PAGE_RESPONSE_MAIN_LANG
        content_for_translated = constants.ARTICLE_PAGE_RESPONSE_FRENCH
        content_copy = dict(content)
        content_for_translated_copy = dict(content_for_translated)

        parent = self.section_index

        self.assertEqual(ArticlePage.objects.count(), 0)

        article = self.importer.create_page(parent, content_copy)

        self.assertEqual(ArticlePage.objects.count(), 1)

        translated_article = self.importer.create_translated_content(
            article, content_for_translated_copy, "fr")

        self.assertEqual(ArticlePage.objects.count(), 2)
        site = Site.objects.get(pk=self.importer.site_pk)
        self.assertEqual(article.get_translation_for(
            "fr", site),
            translated_article)

    def test_create_footer_page(self):
        content = constants.ARTICLE_PAGE_RESPONSE
        content["meta"]["type"] = "core.FooterPage"
        content_copy = dict(content)

        # create local versions of images, mapped to foreign ID
        foreign_image_id = content["image"]["id"]
        image = Image.objects.create(
            title=content["image"]["title"],
            file=get_test_image_file(),
        )
        self.importer.image_map[foreign_image_id] = image.id

        foreign_social_media_image_id = content["social_media_image"]["id"]
        social_media_image = Image.objects.create(
            title=content["social_media_image"]["title"],
            file=get_test_image_file(),
        )

        (self.importer
             .image_map[foreign_social_media_image_id]) = social_media_image.id

        parent = self.footer_index

        self.assertEqual(ArticlePage.objects.count(), 0)

        footer_page = self.importer.create_page(parent, content_copy)

        self.assertEqual(FooterPage.objects.count(), 1)
        self.assertEqual(footer_page.get_parent(), parent)

        self.check_article_and_footer_fields(footer_page, content)

    def test_create_banner_page(self):
        content = constants.BANNER_PAGE_RESPONSE
        content_copy = dict(content)

        foreign_image_id = content["banner"]["id"]
        image = Image.objects.create(
            title=content["banner"]["title"],
            file=get_test_image_file(),
        )
        self.importer.image_map[foreign_image_id] = image.id

        parent = self.banner_index

        self.assertEqual(BannerPage.objects.count(), 0)

        banner_page = self.importer.create_page(parent, content_copy)

        self.assertEqual(BannerPage.objects.count(), 1)
        self.assertEqual(banner_page.get_parent(), parent)

        self.assertEqual(self.importer.id_map[content["id"]], banner_page.id)

        self.assertEqual(banner_page.title, content["title"])
        self.assertEqual(banner_page.external_link, content["external_link"])

        # check that banner link has been created
        self.assertEqual(
            self.importer.banner_page_links[banner_page.id],
            content["banner_link_page"]["id"])

        # check that banner image has been attached
        self.assertTrue(banner_page.banner)
        self.assertEqual(banner_page.banner.title, content["banner"]["title"])

    def test_create_tag_page(self):
        content = constants.TAG_PAGE_RESPONSE
        content_copy = dict(content)

        parent = self.tag_index

        self.assertEqual(Tag.objects.count(), 0)

        tag = self.importer.create_page(parent, content_copy)

        self.assertEqual(Tag.objects.count(), 1)
        self.assertEqual(tag.get_parent(), parent)

        self.assertEqual(self.importer.id_map[content["id"]], tag.id)

        self.assertEqual(tag.title, content["title"])

    def test_create_recommended_articles(self):
        section = self.mk_section(
            self.section_index, title="Parent Test Section 1",
        )
        # create 2 articles
        self.mk_articles(section)
        article_main = ArticlePage.objects.first()
        article_rec = ArticlePage.objects.last()

        # update map_id
        # attach imaginary foreign IDs to articles, to fake import data
        self.importer.id_map = {111: article_main.id, 222: article_rec.id}
        # refer copied page to foreign id of recomended article
        self.importer.recommended_articles[article_main.id] = [222]

        self.assertEqual(ArticlePageRecommendedSections.objects.count(), 0)
        self.importer.create_recommended_articles()
        self.assertEqual(ArticlePageRecommendedSections.objects.count(), 1)
        recomendation = ArticlePageRecommendedSections.objects.first()
        self.assertEqual(recomendation.page.specific, article_main)
        self.assertEqual(recomendation.recommended_article.specific,
                         article_rec)

    def test_create_related_sections(self):
        section_main = self.mk_section(
            self.section_index, title="Parent Test Section 1",
        )
        section_rel = self.mk_section(
            self.section_index, title="Parent Test Section 1",
        )
        # create articles
        self.mk_article(section_main)
        article = ArticlePage.objects.first()

        # update map_id
        # attach imaginary foreign IDs to sections, to fake import data
        self.importer.id_map = {111: section_main.id, 222: section_rel.id}
        # refer copied page to foreign id of related section
        self.importer.related_sections[article.id] = [222]

        self.assertEqual(ArticlePageRelatedSections.objects.count(), 0)
        self.importer.create_related_sections()
        self.assertEqual(ArticlePageRelatedSections.objects.count(), 1)
        relation = ArticlePageRelatedSections.objects.first()
        self.assertEqual(relation.page.specific, article)
        self.assertEqual(relation.section.specific,
                         section_rel)

    def test_create_nav_tag_relationships(self):
        section = self.mk_section(
            self.section_index, title="Parent Test Section 1",
        )
        self.mk_article(section)
        article = ArticlePage.objects.first()

        # create tag
        [tag_1, tag_2] = self.mk_tags(self.tag_index, count=2)

        # update map_id
        # attach imaginary foreign IDs to sections, to fake import data
        self.importer.id_map = {111: tag_1.id, 222: tag_2.id}
        self.importer.nav_tags[article.id] = [111, 222]

        self.assertEqual(ArticlePageTags.objects.count(), 0)
        self.importer.create_nav_tag_relationships()
        self.assertEqual(ArticlePageTags.objects.count(), 2)

        [relation_1, relation_2] = list(ArticlePageTags.objects.all())

        self.assertEqual(relation_1.page.specific, article)
        self.assertEqual(relation_1.tag.specific,
                         tag_1)

        self.assertEqual(relation_2.page.specific, article)
        self.assertEqual(relation_2.tag.specific,
                         tag_2)

    def test_create_section_tag_relationships(self):
        section = self.mk_section(self.section_index)

        # create tag
        [tag_1, tag_2] = self.mk_tags(self.tag_index, count=2)

        # update map_id
        # attach imaginary foreign IDs to sections, to fake import data
        self.importer.id_map = {111: tag_1.id, 222: tag_2.id}
        self.importer.section_tags[section.id] = [111, 222]

        self.assertEqual(SectionPageTags.objects.count(), 0)
        self.importer.create_section_tag_relationship()
        self.assertEqual(SectionPageTags.objects.count(), 2)

        [relation_1, relation_2] = list(SectionPageTags.objects.all())

        self.assertEqual(relation_1.page.specific, section)
        self.assertEqual(relation_1.tag.specific,
                         tag_1)

        self.assertEqual(relation_2.page.specific, section)
        self.assertEqual(relation_2.tag.specific,
                         tag_2)

    def test_create_section_page_translated(self):
        # create 2 languages
        self.english = SiteLanguageRelation.objects.create(
            language_setting=self.importer.language_setting,
            locale='en',
            is_active=True)
        self.french = SiteLanguageRelation.objects.create(
            language_setting=self.importer.language_setting,
            locale='fr',
            is_active=True)

        content = constants.SECTION_PAGE_RESPONSE
        content_for_translated = constants.SECTION_PAGE_RESPONSE_FRENCH
        content_copy = dict(content)
        content_for_translated_copy = dict(content_for_translated)

        parent = self.section_index

        self.assertEqual(SectionPage.objects.count(), 0)

        section = self.importer.create_page(parent, content_copy)

        self.assertEqual(SectionPage.objects.count(), 1)

        translated_section = self.importer.create_translated_content(
            section, content_for_translated_copy, "fr")

        self.assertEqual(SectionPage.objects.count(), 2)
        site = Site.objects.get(pk=self.importer.site_pk)
        self.assertEqual(section.get_translation_for(
            "fr", site),
            translated_section)

    def test_create_banner_page_links(self):
        banner = self.mk_banner(parent=self.banner_index)
        section = self.mk_section(parent=self.section_index)
        # fake the banner page data
        fake_foreign_id = 111
        self.importer.id_map[fake_foreign_id] = section.id
        self.importer.banner_page_links[banner.id] = fake_foreign_id

        self.assertFalse(banner.banner_link_page)

        self.importer.create_banner_page_links()

        banner = BannerPage.objects.get(id=banner.id)
        self.assertTrue(banner.banner_link_page)
        self.assertEqual(banner.banner_link_page.specific, section)

    @patch("molo.core.api.importers.requests.get",
           side_effect=utils.mocked_requests_get)
    def test_get_foreign_page_id_from_type(self, mock_get):
        page_type = "core.SectionIndexPage"
        _id = self.importer.get_foreign_page_id_from_type(page_type)
        self.assertEqual(
            _id,
            constants.TYPE_SECTION_INDEX_PAGE_RESPONSE['items'][0]["id"])

    @patch("molo.core.api.importers.requests.get",
           side_effect=utils.mocked_requests_get)
    def test_recursive_copy(self, mock_get):
        '''
        This test will copy content with the following structure:
        |--Sections
           |--Section 1 [with French translation]
           |  |--Article 1  [with French translation]
           |  |--Article 2
           |
           |--Section 2
              |--Sub Section
                 |--Article 3
        '''
        self.french = SiteLanguageRelation.objects.create(
            language_setting=self.importer.language_setting,
            locale='fr',
            is_active=True)

        self.assertEqual(SectionPage.objects.count(), 0)
        self.assertEqual(ArticlePage.objects.count(), 0)

        self.importer.copy_children(
            foreign_id=constants.SECTION_INDEX_PAGE_RESPONSE["id"],
            existing_node=self.section_index)

        self.assertEqual(SectionPage.objects.count(), 4)
        self.assertEqual(ArticlePage.objects.count(), 4)

        sec_1 = SectionPage.objects.get(
            title=constants.SECTION_RESPONSE_1["title"])
        self.assertTrue(sec_1)
        self.assertEqual(sec_1.get_parent().specific, self.section_index)
        self.assertEqual(sec_1.get_children().count(), 3)

        sec_1_trans = SectionPage.objects.get(
            title=constants.SECTION_RESPONSE_1_TRANSLATION_1["title"])
        self.assertTrue(sec_1_trans)
        self.assertEqual(sec_1_trans.get_parent().specific, self.section_index)
        self.assertEqual(sec_1_trans.get_children().count(), 0)

        sec_2 = SectionPage.objects.get(
            title=constants.SECTION_RESPONSE_2["title"])
        self.assertTrue(sec_2)
        self.assertEqual(sec_2.get_parent().specific, self.section_index)
        self.assertEqual(sec_2.get_children().count(), 1)

        sec_3 = SectionPage.objects.get(
            title=constants.SUB_SECTION_RESPONSE_1["title"])
        self.assertTrue(sec_3)
        self.assertEqual(sec_3.get_parent().specific, sec_2)
        self.assertEqual(sec_3.get_children().count(), 1)

        art_1 = ArticlePage.objects.get(
            title=constants.ARTICLE_RESPONSE_1["title"])
        self.assertTrue(art_1)
        self.assertTrue(art_1.get_parent().specific, sec_1)

        art_1_trans = ArticlePage.objects.get(
            title=constants.ARTICLE_RESPONSE_1_TRANSLATION["title"])
        self.assertTrue(art_1_trans)
        self.assertTrue(art_1_trans.get_parent().specific, sec_1)

        art_2 = ArticlePage.objects.get(
            title=constants.ARTICLE_RESPONSE_2["title"])
        self.assertTrue(art_2)
        self.assertTrue(art_2.get_parent().specific, sec_1)

        art_3 = ArticlePage.objects.get(
            title=constants.NESTED_ARTICLE_RESPONSE["title"])
        self.assertTrue(art_3)
        self.assertTrue(art_3.get_parent().specific, sec_3)



import os
import shutil
import zipfile
import re
import tempfile
import imagehash
import distutils.dir_util

from PIL import Image as PILImage
from StringIO import StringIO
from django.conf import settings
from wagtail.wagtailcore.utils import cautious_slugify
from wagtail.wagtailcore.models import Page


def create_new_article_relations(old_main, copied_main):
    from molo.core.models import ArticlePage, Tag, ArticlePageTags, \
        ArticlePageReactionQuestions, ReactionQuestion, \
        ArticlePageRecommendedSections, ArticlePageRelatedSections, \
        SectionPage, BannerPage
    if old_main and copied_main:
        if copied_main.get_descendants().count() >= \
                old_main.get_descendants().count():
            for article in ArticlePage.objects.descendant_of(copied_main):

                # replace old tag with new tag in tag relations
                tag_relations = ArticlePageTags.objects.filter(page=article)
                for relation in tag_relations:
                    if relation.tag:
                        new_tag = Tag.objects.descendant_of(
                            copied_main).filter(slug=relation.tag.slug).first()
                        relation.tag = new_tag
                        relation.save()
                # replace old reaction question with new reaction question
                question_relations = \
                    ArticlePageReactionQuestions.objects.filter(page=article)
                for relation in question_relations:
                    if relation.reaction_question:
                        new_question = ReactionQuestion.objects.descendant_of(
                            copied_main).filter(
                                slug=relation.reaction_question.slug).first()
                        relation.reaction_question = new_question
                        relation.save()

                # replace old recommended articles with new articles
                recommended_article_relations = \
                    ArticlePageRecommendedSections.objects.filter(page=article)
                for relation in recommended_article_relations:
                    if relation.recommended_article:
                        new_recommended_article = \
                            ArticlePage.objects.descendant_of(
                                copied_main).filter(
                                slug=relation.recommended_article.slug).first()
                        relation.recommended_article = new_recommended_article
                        relation.save()

                # replace old related sections with new sections
                related_section_relations = \
                    ArticlePageRelatedSections.objects.filter(page=article)
                for relation in related_section_relations:
                    if relation.section:
                        new_related_section = \
                            SectionPage.objects.descendant_of(
                                copied_main).filter(
                                slug=relation.section.slug).first()
                        relation.section = new_related_section
                        relation.save()

                # replace old article banner relations with new articles
                for banner in BannerPage.objects.descendant_of(copied_main):
                    old_page = Page.objects.get(
                        pk=banner.banner_link_page.pk)
                    if old_page:
                        new_article = Page.objects.descendant_of(
                            copied_main).get(slug=old_page.slug)
                        banner.banner_link_page = new_article
                        banner.save()


def get_locale_code(language_code=None):
    return (language_code or settings.LANGUAGE_CODE).replace('_', '-')


RE_NUMERICAL_SUFFIX = re.compile(r'^[\w-]*-(\d+)+$')


def generate_slug(text, tail_number=0):
    from wagtail.wagtailcore.models import Page
    """
    Returns a new unique slug. Object must provide a SlugField called slug.
    URL friendly slugs are generated using django.template.defaultfilters'
    slugify. Numbers are added to the end of slugs for uniqueness.

    based on implementation in jmbo.utils
    https://github.com/praekelt/jmbo/blob/develop/jmbo/utils/__init__.py
    """

    # Empty slugs are ugly (eg. '-1' may be generated) so force non-empty
    if not text:
        text = 'no-title'

    # use django slugify filter to slugify
    slug = cautious_slugify(text)[:255]

    values_list = Page.objects.filter(
        slug__startswith=slug
    ).values_list('id', 'slug')

    # Find highest suffix
    max = -1
    for tu in values_list:
        if tu[1] == slug:
            if max == -1:
                # Set max to indicate a collision
                max = 0

        # Update max if suffix is greater
        match = RE_NUMERICAL_SUFFIX.match(tu[1])
        if match is not None:
            i = int(match.group(1))
            if i > max:
                max = i

    if max >= 0:
        # There were collisions
        return "%s-%s" % (slug, max + 1)
    else:
        # No collisions
        return slug


def update_media_file(upload_file):
    '''
    Update the Current Media Folder.

    Returns list of files copied across or
    raises an exception.
    '''
    temp_directory = tempfile.mkdtemp()
    temp_file = tempfile.TemporaryFile()
    # assumes the zip file contains a directory called media
    temp_media_file = os.path.join(temp_directory, 'media')
    try:
        for chunk in upload_file.chunks():
            temp_file.write(chunk)

        with zipfile.ZipFile(temp_file, 'r') as z:
            z.extractall(temp_directory)

        if os.path.exists(temp_media_file):
            return distutils.dir_util.copy_tree(
                temp_media_file,
                settings.MEDIA_ROOT)
        else:
            raise Exception("Error: There is no directory called "
                            "'media' in the root of the zipped file")
    except Exception as e:
        raise e
    finally:
        temp_file.close()
        if os.path.exists(temp_directory):
            shutil.rmtree(temp_directory)


def get_image_hash(image):
    '''
    Returns an image hash of a Wagtail Image
    '''
    with open(image.file.path, 'r') as file:
        image_in_memory = StringIO(file.read())
        pil_image = PILImage.open(image_in_memory)
        return imagehash.average_hash(pil_image).__str__()



from collections import OrderedDict

from rest_framework import serializers

from wagtail.api.v2.serializers import PageSerializer
from wagtail.wagtailimages.api.v2.serializers import ImageSerializer

from molo.core.models import TranslatablePageMixinNotRoutable
from molo.core.utils import get_image_hash


class PageChildrenField(serializers.Field):
    """
    Serializes the "children" field.

    Example:
    "children": {
        "count": 1,
        "listing_url": "/api/v1/pages/?child_of=2"
    }
    """
    def get_attribute(self, instance):
        return instance

    def to_representation(self, page):
        if page.get_children():
            items = []
            for item in page.get_children():
                items.append(item.id)
            return OrderedDict([
                ("items", items),
            ])


class MainLanguageChildrenField(serializers.Field):
    """
    Serializes the "children" field with only children that belong
    to the main language.

    Example:
    "children": [12, 34]
    """
    def get_attribute(self, instance):
        return instance

    def to_representation(self, page):
        if page.get_children():
            return [child_page.id for child_page
                    in page.get_children()
                    if not hasattr(child_page.specific, 'source_page')]


class PageTranslationsField(serializers.Field):
    """
    Serializes the "translations" field.

    If the Page is translatable, it will return all of the translated
    versions of the page, including the main lanugage page. It will
    exclude the original page id from the list.

    Example:
    "translations": [
        {
            "locale": "ve",
            "id": 24
        },
        {
            "locale": "fr",
            "id": 25
        }
    ]

    """
    def get_attribute(self, instance):
        return instance

    def to_representation(self, page):
        if isinstance(page.specific, TranslatablePageMixinNotRoutable):
            main_page = page.specific.get_main_language_page().specific
            items = []

            if page.specific == main_page:
                pages = [page_relation.translated_page.specific
                         for page_relation in main_page.translations.all()]
            else:
                pages = ([main_page] +
                         [page_relation.translated_page.specific
                         for page_relation in main_page.translations.all()
                         if page_relation.translated_page.id != page.id])

            for translated_page in pages:
                items.append({
                    "id": translated_page.id,
                    "locale": translated_page.languages.get().language.locale
                })
            return items
        else:
            return None


class MoloPageSerializer(PageSerializer):
    children = PageChildrenField(read_only=True)
    translations = PageTranslationsField(read_only=True)
    main_language_children = MainLanguageChildrenField(read_only=True)


class ImageUrlField(serializers.Field):
    """
    Provides an accesible URL for image file
    """
    def get_attribute(self, instance):
        return instance

    def to_representation(self, image):
        rendition = image.get_rendition('original')
        return rendition.url


class ImageFileHashField(serializers.Field):
    """
    Provides an accesible URL for image file
    """
    def get_attribute(self, instance):
        return instance

    def to_representation(self, image):
        return get_image_hash(image)


class MoloImageSerializer(ImageSerializer):
    image_url = ImageUrlField(read_only=True)
    image_hash = ImageFileHashField(read_only=True)



"""
Various importers for the different content types
"""
import math
import json
import requests
from io import BytesIO

from django.core.files.images import ImageFile

from wagtail.wagtailcore.models import Page
from wagtail.wagtailimages.models import Image

from molo.core.models import (
    Languages,
    SiteLanguageRelation,
    ArticlePage,
    SectionPage,
    FooterPage,
    BannerPage,
    Tag,
    ArticlePageRecommendedSections,
    ArticlePageRelatedSections,
    PageTranslation,
    ArticlePageTags,
    SectionPageTags,
)
from molo.core.api.constants import (
    API_IMAGES_ENDPOINT, API_PAGES_ENDPOINT, KEYS_TO_EXCLUDE,
)
from molo.core.utils import get_image_hash

from django.core.exceptions import ObjectDoesNotExist


# functions used to find images
def get_image_attributes(base_url, image_id):
    image_attrs = requests.get(
        base_url + API_IMAGES_ENDPOINT + str(image_id)
    ).json()
    return image_attrs


def get_image(base_url, image_id):
    # TODO: guard against non-existent images
    image_attributes = get_image_attributes(base_url, image_id)
    image_file = requests.get(image_attributes["file"])
    image = Image(
        title=image_attributes["title"],
        file=ImageFile(
            BytesIO(image_file.content), name=image_attributes["title"]
        )
    )
    image.save()
    return image


def separate_fields(fields):
    """
    Non-foreign key fields can be mapped to new article instances
    directly. Foreign key fields require a bit more work.
    This method returns a tuple, of the same format:
    (flat fields, nested fields)
    """
    flat_fields = {}
    nested_fields = {}

    # exclude "id" and "meta" elements
    for k, v in fields.items():
        if k not in KEYS_TO_EXCLUDE:
            if type(v) not in [type({}), type([])]:
                flat_fields.update({k: v})
            else:
                nested_fields.update({k: v})

    return flat_fields, nested_fields


def list_of_objects_from_api(url):
        '''
        API only serves 20 pages by default
        This fetches info on all of items and return them as a list

        Assumption: limit of API is not less than 20
        '''
        response = requests.get(url)

        content = json.loads(response.content)
        count = content["meta"]["total_count"]

        if count <= 20:
            return content["items"]
        else:
            items = [] + content["items"]
            num_requests = int(math.ceil(count // 20))

            for i in range(1, num_requests + 1):
                paginated_url = "{}?limit=20&offset={}".format(
                    url, str(i * 20))
                paginated_response = requests.get(paginated_url)
                items = items + json.loads(paginated_response.content)["items"]
        return items


def record_foreign_relation(field, key, record_keeper, id_key="id"):
    '''
    returns a function with the attributes necessary to
    correctly reference the necessary objects, to correctly record
    the relationship between a page and its foreign foreign-key pages
    '''
    def record_relationship(nested_fields, page_id):
        if ((field in nested_fields) and nested_fields[field]):
            record_keeper[page_id] = []
            for thing in nested_fields[field]:
                record_keeper[page_id].append(thing[key][id_key])
    return record_relationship


def record_foreign_key(field, record_keeper, id_key="id"):
    '''
    returns a function with the attributes necessary to replicate
    a foreign key relation
    '''
    def _record_foreign_key(nested_fields, page_id):
        if ((field in nested_fields) and nested_fields[field]):
            record_keeper[page_id] = nested_fields[field][id_key]
    return _record_foreign_key


def add_json_dump(field):
    def _add_json_dump(nested_fields, page):
        if ((field in nested_fields) and
                nested_fields[field]):
            setattr(page, field, json.dumps(nested_fields[field]))
    return _add_json_dump


def add_list_of_things(field):
    def _add_list_of_things(nested_fields, page):
        if (field in nested_fields) and nested_fields[field]:
            attr = getattr(page, field)
            for item in nested_fields[field]:
                attr.add(item)
    return _add_list_of_things


def attach_image(field, image_map):
    '''
    Returns a function that attaches an image to page if it exists

    Assumes that images have already been imported
    otherwise will fail silently
    '''
    def _attach_image(nested_fields, page):
        if (field in nested_fields) and nested_fields[field]:
            foreign_image_id = nested_fields[field]["id"]
            try:
                local_image_id = image_map[foreign_image_id]
                local_image = Image.objects.get(id=local_image_id)
                setattr(page, field, local_image)
            except (KeyError, ObjectDoesNotExist):
                # TODO: log when page is not found
                pass
    return _attach_image


class PageImporter(object):

    def __init__(self, base_url=None, content=None, content_type=None):
        self._content_type = content_type
        self._fields = []
        self._content = content
        self._base_url = base_url

    def get_content_from_url(self, base_url):
        """
        Sections can have SectionPage and ArticlePage child objects.
        These have different fields, and thus have to be treated
        differently.
        """
        # assemble url
        base_url = base_url.rstrip("/")
        url = base_url + API_PAGES_ENDPOINT + "?type=" + self._content_type + \
            "&fields=" + ",".join(self._fields) + \
            "&order=latest_revision_created_at"

        # make request
        try:
            response = requests.get(url)
            self._base_url = base_url
            self._content = response.json()
            self._content = self._content["items"]
            return self._content
        except requests.exceptions.ConnectionError:
            return "No content could be found from {}. " \
                "Are you sure this is the correct URL?".format(base_url)
        except requests.exceptions.RequestException:
            return "Content could not be imported at this time. " \
                   "Please try again later."

    def content(self):
        if self._content:
            return self._content
        return []

    def save(self, indexes, parent_id):
        pass


class ArticlePageImporter(PageImporter):

    def __init__(self, base_url=None, content=None, content_type=None):
        super(ArticlePageImporter, self).__init__(
            base_url=base_url, content=content, content_type=content_type
        )
        self._content_type = "core.ArticlePage"
        self._fields = ArticlePage.get_api_fields()

    def save(self, indexes, parent_id):
        if self.content():
            parent = Page.objects.get(id=parent_id)
            for index in indexes:
                # Remove "id" and "meta" fields
                selected_article = self.content()[index]
                fields, nested_fields = separate_fields(selected_article)
                article = ArticlePage(**fields)

                # TODO: u'related_sections'
                # process the nested fields
                if ("tags" in nested_fields) and nested_fields["tags"]:
                    article.tags.add(", ".join(nested_fields["tags"]))

                if ("metadata_tags" in nested_fields) and \
                        nested_fields["metadata_tags"]:
                    article.metadata_tags.add(
                        ", ".join(nested_fields["metadata_tags"])
                    )

                if ("body" in nested_fields) and nested_fields["body"]:
                    article.body = json.dumps(nested_fields["body"])

                if ("image" in nested_fields) and nested_fields["image"]:
                    article.image = get_image(
                        self._base_url, nested_fields["image"]["id"]
                    )

                parent.add_child(instance=article)
                parent.save_revision().publish()


class SectionPageImporter(PageImporter):

    def __init__(self, base_url=None, content=None, content_type=None):
        super(SectionPageImporter, self).__init__(
            base_url=base_url, content=content, content_type=content_type
        )
        self._content_type = "core.SectionPage"
        self._fields = SectionPage.get_api_fields()

    def _save_item(self, item, parent):
        if item["meta"]["type"] == "core.SectionPage":
            flat_fields, nested_fields = separate_fields(item)
            child_section = SectionPage(**flat_fields)

            if ("image" in nested_fields) and nested_fields["image"]:
                child_section.image = get_image(
                    self._base_url, nested_fields["image"]["id"]
                )

            parent.add_child(instance=child_section)
            parent.save_revision().publish()
            return child_section

        elif item["meta"]["type"] == "core.ArticlePage":
            flat_fields, nested_fields = separate_fields(item)
            child_article = ArticlePage(**flat_fields)
            if ("tags" in nested_fields) and nested_fields["tags"]:
                child_article.tags.add(
                    ", ".join(nested_fields["tags"])
                )

            if ("metadata_tags" in nested_fields) and \
                    nested_fields["metadata_tags"]:
                child_article.metadata_tags.add(
                    ", ".join(nested_fields["metadata_tags"])
                )

            if ("body" in nested_fields) and nested_fields["body"]:
                child_article.body = json.dumps(nested_fields["body"])

            if ("image" in nested_fields) and nested_fields["image"]:
                child_article.image = get_image(
                    self._base_url, nested_fields["image"]["id"]
                )

            parent.add_child(instance=child_article)
            parent.save_revision().publish()
            return child_article

    def process_child_section(self, id, parent):
        response = requests.get(
            self._base_url + API_PAGES_ENDPOINT + str(id)
        ).json()
        ancestor = self._save_item(response, parent)
        if response["meta"]["children"]:
            for child_id in response["meta"]["children"]["items"]:
                self.process_child_section(child_id, ancestor)
        return ancestor

    def recursive_children(self, node):
        results = [node['id']]
        if len(node['children']) > 0:
            for child in node['children']:
                results.extend(self.recursive_children(child))
        return results

    def save(self, indexes, parent_id):
        """
        Save the selected section. This will save the selected section
        as well as its direct child pages obtained through the ?child_of
        query parameter. The ?descendant_of query parameter is probably
         better suited because it all pages under that part of the tree will
         be obtained. The problem , however, is that that will require being
         able to traverse the tree and recreate parent-child relationships
         after they are imported
        """
        if self.content():
            parent = Page.objects.get(id=parent_id)

            # Save the selected section page
            response = requests.get(
                self._base_url + API_PAGES_ENDPOINT + str(indexes[0]) + "/"
            )

            section_page = response.json()
            self.process_child_section(section_page["id"], parent)


class SiteImporter(object):

    def __init__(self, site_pk, base_url=''):
        if base_url[-1] == '/':
            self.base_url = base_url[:-1]
        else:
            self.base_url = base_url

        self.api_url = self.base_url + '/api/v2/'
        self.image_url = "{}images/".format(self.api_url)
        self.site_pk = site_pk
        self.language_setting, created = Languages.objects.get_or_create(
            site_id=self.site_pk)
        self.content = None
        # maps foreign IDs to local page IDs
        self.id_map = {}
        # maps foreign image IDs to local IDs
        self.image_map = {}
        # maps local id to list of foreign section page ids
        self.related_sections = {}
        # maps local id to list of foreign page ids
        self.recommended_articles = {}
        # maps local pages to list of foreign reaction question IDs
        self.reaction_questions = {}
        # maps local pages to list of foreign nav_tag IDs
        self.nav_tags = {}
        # maps local pages to list of foreign section_tag IDs
        self.section_tags = {}
        # maps local banner page id to foreign linked page id
        self.banner_page_links = {}

        self.record_recommended_articles = record_foreign_relation(
            "recommended_articles", "recommended_article",
            self.recommended_articles)
        self.record_section_tags = record_foreign_relation(
            "section_tags", "tag",
            self.section_tags)
        self.record_nav_tags = record_foreign_relation(
            "nav_tags", "tag",
            self.nav_tags)
        self.record_reaction_questions = record_foreign_relation(
            "reaction_questions", "reaction_question",
            self.reaction_questions)
        self.record_related_sections = record_foreign_relation(
            "related_sections", "section",
            self.related_sections)
        self.record_banner_page_link = record_foreign_key(
            "banner_link_page",
            self.banner_page_links)

        self.add_article_body = add_json_dump("body")
        self.add_section_time = add_json_dump("time")

        self.add_tags = add_list_of_things("tags")
        self.add_metadata_tags = add_list_of_things("metadata_tags")

        self.attach_image = attach_image("image", self.image_map)
        self.attach_social_media_image = attach_image("social_media_image",
                                                      self.image_map)
        self.attach_banner_image = attach_image("banner", self.image_map)

    def get_language_ids(self):
        language_url = "{}{}/".format(self.api_url, "languages")
        response = requests.get(language_url)

        language_ids = []
        # TODO: handle broken case
        for language in json.loads(response.content)["items"]:
            language_ids.append(language["id"])
        return language_ids

    def copy_site_languages(self):
        language_foreign_ids = self.get_language_ids()
        language_setting, created = Languages.objects.get_or_create(
            site_id=self.site_pk)
        for foreign_id in language_foreign_ids:
            language_url = "{}{}/{}/".format(self.api_url,
                                             "languages",
                                             foreign_id)
            response = requests.get(language_url)
            content = json.loads(response.content)

            # TODO: review whether this works with Multi-site
            # TODO: handle case where Main lang is not first
            sle, created = SiteLanguageRelation.objects.get_or_create(
                locale=content['locale'],
                is_active=content['is_active'],
                language_setting=language_setting)

    def import_images(self):
        '''
        Fetches images from site

        Attempts to avoid duplicates by matching image titles
        if a match is found it refers to local instance instead
        if it is not, the image is fetched, created and referenced
        '''
        images = list_of_objects_from_api(self.image_url)

        if not images:
            return None

        # store info about local images in order to match
        # with imported images
        local_image_hashes = {}
        image_width = {}
        image_height = {}

        for local_image in Image.objects.all():
            local_image_hashes[get_image_hash(local_image)] = local_image

            if local_image.width in image_width:
                image_width[local_image.width].append(local_image)
            else:
                image_width[local_image.width] = [local_image]

            if local_image.height in image_height:
                image_height[local_image.height].append(local_image)
            else:
                image_height[local_image.height] = [local_image]

        # iterate through foreign images
        for image in images:
            image_detail_url = "{}{}/".format(self.image_url, image["id"])
            img_response = requests.get(image_detail_url)
            img_info = json.loads(img_response.content)

            if img_info["image_hash"] is None:
                raise ValueError('image hash should not be none')

            # check if a replica exists
            if img_info["width"] in image_width:
                possible_matches = image_width[img_info["width"]]
                if img_info["height"] in image_height:
                    possible_matches = list(
                        set(image_height[img_info["height"]] +
                            possible_matches))
                    if img_info["image_hash"] in local_image_hashes:
                        result = list(
                            set([local_image_hashes[img_info["image_hash"]]] +
                                possible_matches))
                        # edge case where we have more than one match
                        result = result[0]

                        # use the local title of the image
                        self.image_map[image["id"]] = result.id
                        # LOG THIS
                        continue

            new_image = self.fetch_and_create_image(
                img_info['image_url'],
                img_info["title"])
            self.image_map[image["id"]] = new_image.id

    def fetch_and_create_image(self, relative_url, image_title):
        '''
        fetches, creates and return image object
        '''
        # TODO: handle image unavailable
        image_media_url = "{}{}".format(self.base_url, relative_url)
        image_file = requests.get(image_media_url)
        local_image = Image(
            title=image_title,
            file=ImageFile(
                BytesIO(image_file.content), name=image_title
            )
        )
        local_image.save()
        return local_image

    def create_translated_content(self, local_main_lang_page,
                                  content, locale):
        '''
        Wrapper for  create_content

        Creates the content
        Then attaches a language relation from the main language page to the
        newly created Page
        Note: we get the parent from the main language page
        '''
        page = self.create_page(local_main_lang_page.get_parent(), content)

        language = SiteLanguageRelation.objects.get(
            language_setting=self.language_setting, locale=locale)
        language_relation = page.languages.first()
        language_relation.language = language
        language_relation.save()
        page.save_revision().publish()
        PageTranslation.objects.get_or_create(
            page=local_main_lang_page,
            translated_page=page)
        return page

    def create_page(self, parent, content):
        fields, nested_fields = separate_fields(content)

        # handle the unwanted fields
        foreign_id = content.pop('id')
        # ignore when article was last revised
        if 'latest_revision_created_at' in content:
            content.pop('latest_revision_created_at')

        page = None
        if content["meta"]["type"] == "core.SectionPage":
            page = SectionPage(**fields)
        elif content["meta"]["type"] == "core.ArticlePage":
            page = ArticlePage(**fields)
        elif content["meta"]["type"] == "core.FooterPage":
            page = FooterPage(**fields)
        elif content["meta"]["type"] == "core.BannerPage":
            page = BannerPage(**fields)
        elif content["meta"]["type"] == "core.Tag":
            page = Tag(**fields)

        # TODO: handle other Page types

        parent.add_child(instance=page)

        # TODO: handle live/published
        # Need to review this line:
        #   handle drafts and 'not live' content
        parent.save_revision().publish()

        self.id_map[foreign_id] = page.id

        self.record_section_tags(nested_fields, page.id)
        self.record_nav_tags(nested_fields, page.id)
        self.record_reaction_questions(nested_fields, page.id)
        self.record_recommended_articles(nested_fields, page.id)
        self.record_related_sections(nested_fields, page.id)
        self.record_banner_page_link(nested_fields, page.id)

        self.add_article_body(nested_fields, page)
        self.add_section_time(nested_fields, page)

        self.add_tags(nested_fields, page)
        self.add_metadata_tags(nested_fields, page)

        self.attach_image(nested_fields, page)
        self.attach_social_media_image(nested_fields, page)
        self.attach_banner_image(nested_fields, page)

        # note that unpublished pages will be published
        page.save_revision().publish()
        return page

    def create_recommended_articles(self):
        # iterate through articles with recomended articles
        for article_id, foreign_rec_article_id_list in self.recommended_articles.iteritems():  # noqa

            main_article = ArticlePage.objects.get(id=article_id)
            for foreign_rec_article_id in foreign_rec_article_id_list:
                local_version_page_id = self.id_map[foreign_rec_article_id]
                rec_article = ArticlePage.objects.get(id=local_version_page_id)

                ArticlePageRecommendedSections(
                    page=main_article,
                    recommended_article=rec_article
                ).save()

    def create_related_sections(self):
        # iterate through articles with related sections
        for article_id, foreign_rel_section_id_list in self.related_sections.iteritems():  # noqa

            main_article = ArticlePage.objects.get(id=article_id)
            for foreign_rel_section_id in foreign_rel_section_id_list:
                local_version_page_id = self.id_map[foreign_rel_section_id]
                rel_section = SectionPage.objects.get(id=local_version_page_id)

                ArticlePageRelatedSections(
                    page=main_article,
                    section=rel_section
                ).save()

    def create_nav_tag_relationships(self):
        for page_id, foreign_tags_id_list in self.nav_tags.iteritems():
            page = Page.objects.get(id=page_id).specific
            for foreign_tag_id in foreign_tags_id_list:
                local_tag_id = self.id_map[foreign_tag_id]
                tag = Tag.objects.get(id=local_tag_id)

                ArticlePageTags(
                    page=page,
                    tag=tag
                ).save()

    def create_section_tag_relationship(self):
        for page_id, foreign_tags_id_list in self.section_tags.iteritems():
            page = Page.objects.get(id=page_id).specific
            for foreign_tag_id in foreign_tags_id_list:
                local_tag_id = self.id_map[foreign_tag_id]
                tag = Tag.objects.get(id=local_tag_id)

                SectionPageTags(
                    page=page,
                    tag=tag
                ).save()

    def create_banner_page_links(self):
        for banner_page_id, linked_page_foreign_id in self.banner_page_links.iteritems():  # noqa
            banner = BannerPage.objects.get(id=banner_page_id)
            local_id = self.id_map[linked_page_foreign_id]
            linked_page = Page.objects.get(id=local_id).specific
            banner.banner_link_page = linked_page
            banner.save_revision().publish()

    def get_foreign_page_id_from_type(self, page_type):
        '''
        Get the foreign page id based on type

        Only works for index pages
        '''
        response = requests.get("{}pages/?type={}".format(
            self.api_url, page_type))
        content = json.loads(response.content)
        return content["items"][0]["id"]

    def copy_children(self, foreign_id, existing_node):
        '''
        Initiates copying of tree, with existing_node acting as root
        '''
        url = "{}/api/v2/pages/{}/".format(self.base_url, foreign_id)

        response = requests.get(url)
        content = json.loads(response.content)

        main_language_child_ids = content["meta"]["main_language_children"]
        for main_language_child_id in main_language_child_ids:
                self.copy_page_and_children(foreign_id=main_language_child_id,
                                            parent_id=existing_node.id)

    def copy_page_and_children(self, foreign_id, parent_id):
        '''
        Recusively copies over pages, their translations and child pages
        '''
        url = "{}/api/v2/pages/{}/".format(self.base_url, foreign_id)

        # TODO handle connection errors
        response = requests.get(url)
        content = json.loads(response.content)

        parent = Page.objects.get(id=parent_id).specific
        page = self.create_page(parent, content)

        # create translations
        if content["meta"]["translations"]:
            for translation_obj in content["meta"]["translations"]:
                _url = "{}/api/v2/pages/{}/".format(self.base_url,
                                                    translation_obj["id"])
                _response = requests.get(_url)
                _content = json.loads(_response.content)

                self.create_translated_content(
                    page, _content, translation_obj["locale"])

        main_language_child_ids = content["meta"]["main_language_children"]

        # recursively iterate through child nodes
        if main_language_child_ids:
            for main_language_child_id in main_language_child_ids:
                self.copy_page_and_children(foreign_id=main_language_child_id,
                                            parent_id=page.id)



from wagtail.wagtailcore.models import Page
from wagtail.wagtailcore.utils import resolve_model_string
from wagtail.api.v2.endpoints import PagesAPIEndpoint
from wagtail.api.v2.serializers import BaseSerializer
from wagtail.api.v2.utils import BadRequestError
from wagtail.wagtailimages.api.v2.endpoints import ImagesAPIEndpoint
from wagtail.wagtailimages.api.v2.endpoints import BaseAPIEndpoint

from molo.core.models import SiteLanguage
from molo.core.api.filters import MainLanguageFilter
from molo.core.api.serializers import (
    MoloPageSerializer,
    MoloImageSerializer,
)


class MoloImagesAPIEndpoint(ImagesAPIEndpoint):
    base_serializer_class = MoloImageSerializer
    body_fields = ImagesAPIEndpoint.body_fields + [
        "filename",
        "file",
        "image_url",
        "image_hash",
    ]


class MoloPagesEndpoint(PagesAPIEndpoint):
    base_serializer_class = MoloPageSerializer

    meta_fields = PagesAPIEndpoint.meta_fields + [
        "children",
        "translations",
        "main_language_children",
    ]

    filter_backends = PagesAPIEndpoint.filter_backends + [
        MainLanguageFilter,
    ]

    known_query_parameters = PagesAPIEndpoint.known_query_parameters.union([
        'is_main_language'
    ])
    extra_api_fields = ['url', 'live']

    def get_queryset(self):
        '''
        This is overwritten in order to not exclude drafts
        and pages submitted for moderation
        '''
        request = self.request

        # Allow pages to be filtered to a specific type
        if 'type' not in request.GET:
            model = Page
        else:
            model_name = request.GET['type']
            try:
                model = resolve_model_string(model_name)
            except LookupError:
                raise BadRequestError("type doesn't exist")
            if not issubclass(model, Page):
                raise BadRequestError("type doesn't exist")

        # This is the overwritten line
        queryset = model.objects.public()  # exclude .live()

        # Filter by site
        queryset = queryset.descendant_of(
            request.site.root_page, inclusive=True)

        return queryset


class LanguagesAPIEndpoint(BaseAPIEndpoint):
    base_serializer_class = BaseSerializer
    filter_backends = []
    extra_api_fields = []
    name = 'languages'
    model = SiteLanguage
